import random

__author__ = "Pablo Gay"
__copyright__ = "Copyright 2007, Pablo Gay"
__license__ = "GPL"

class Chromosome:
    '''
    Base class for genome. Defines the 3 main functionalities every
    genome must have: crossover, mutation and fitness.
    '''    
    def mutate(self, mutate_prob):
        raise Exception('Genome.mutate not implemented.')
    
    def crossover(self, genome):
        raise Exception('Genome.crossover not implemented.')
    
    def fitness(self, function):
        raise Exception('Genome.fitness not implemented.')
        
class BinaryChromosome(Chromosome):
    '''
    Chromosome class where each gene is a binary value. Parameters are
    represented as sets of bits. 
    '''
    def __init__(self, npar, bits, lo, hi, genes=None):
        self.lo   = lo
        self.hi   = hi
        self.npar = npar
        self.bits = bits
        if genes == None:
            self.genes = [ random.choice([0, 1]) for _ in range(npar * bits) ]
        else:
            self.genes = genes
    
    def mutate(self, mutate_prob):
        '''
        Mutates the genes acording to the mutate_prob probability pased as
        parameter. All genes in the chromosome are tested for mutation.
        '''
        for i in range(self.npar * self.bits):
            if random.random() < mutate_prob:
                self.genes[i] = 1 if self.genes[i] == 0 else 0
    
    def crossover(self, genome):
        '''
        Mates the actual chromosome with the one passed by parameter using a 
        random random position as a single crossover point.
        '''
        idx = int(random.random()*len(self.genes))
        newgenes = self.genes[:idx] + genome.genes[idx:]
        return BinaryChromosome(self.npar, self.bits, self.lo, self.hi, genes=newgenes)
    
    def fitness(self, function):
        '''
        Given a fitness function, decodes de genes and evaluates the
        function with the parameters stored in the genetic information.
        '''
        return function(*self.decode())
    
    def decode(self):
        params = []
        pmax = int('1'*self.bits, base=2)
        for idx in range(0, self.npar*self.bits, self.bits):
            p = int(''.join([str(x) for x in self.genes[idx:idx+self.bits]]), base=2)
            q = p/pmax
            q = q * (self.hi-self.lo) + self.lo
            params.append(q)
        return params
    
    def decode1(self):
        params = []
        pmax = int('1'*self.bits, base=2)
        for idx in range(0, self.npar*self.bits):
            params.append(self.genes[idx])
        return params
    
    def copy(self):
        return BinaryChromosome(self.npar, self.bits, self.lo, self.hi, genes=[x for x in self.genes])
    
    def __str__(self):
#        return 'BC:[{}]'.format(', '.join([str(x) for x in self.decode()]))
         return 'BC:[{}]'.format(', '.join([str(x) for x in self.decode1()]))
    
class RealChromosome(Chromosome):
    '''
    Chromosome class where each gene is a real number representing a parameter.
    '''
    def __init__(self, size, lo, hi, genes=None):
        '''
        RealChromosome constructor. As parameters it requires:
         - size: number of genes that the genome must have
         - lo: lower boundary of the rang of numbers each gene can have.
         - hi: higher boundary of the rang of numbers each gene can have.
         - genes (optional): array with the genes definded by the user.
        '''
        self.lo    = lo
        self.hi    = hi
        self.size  = size
        if genes == None:
            self.genes = [ (random.random() * (hi-lo)) + lo for _ in range(size)]
        else:
            self.genes = genes
    
    def mutate(self, mutate_prob):
        '''
        Mutates the genes acording to the mutate_prob probability pased as
        parameter. All genes in the chromosome are tested for mutation.
        '''
        for i in range(self.size):
            if random.random() < mutate_prob:
                self.genes[i] = (random.random() * (self.hi-self.lo)) + self.lo
    
    def crossover(self, genome):
        '''
        Mates the actual chromosome with the one passed by parameter using a 
        random random position as a single crossover point.
        '''
        idx = int(random.random()*self.size)
        newgenes = self.genes[:idx] + genome.genes[idx:]
        return RealChromosome(self.size, self.lo, self.hi, genes=newgenes)
    
    def fitness(self, function):
        '''
        Given a fitness function, evaluates the function with the
        parameters stored in the genetic information.
        '''
        return function(*self.genes)
    
    def copy(self):
        return RealChromosome(self.size, self.lo, self.hi, genes=[x for x in self.genes])
    
    def __str__(self):
        '''
        Returns the string representation of the chromosome.
        '''
        return 'RC:[{}]'.format(', '.join([str(x) for x in self.genes]))

class Population:
    def __init__(self, retain_prob, mutation_prob, fitness, maximize=True):
        self.retain_prob   = retain_prob
        self.mutation_prob = mutation_prob
        self.fitness       = fitness
        self.maximize      = maximize
        self.population    = []
    
    def addChromosome(self, ch):
        self.population.append(ch)
    
    def evolve(self, max_eras, target_fitness=None, verbose=True):
        info = []
        for era in range(max_eras):
            
            # Sort population
            rank = [(g, g.fitness(self.fitness)) for g in self.population]
            rank.sort(reverse=self.maximize, key=lambda x:x[1])
            self.population = [x[0] for x in rank]
            
            best_gene    = rank[0][0]
            best_fitness = best_gene.fitness(self.fitness)
            avg_fitness  = sum([x[1] for x in rank])/len(self.population)
            
            info.append({'best_genes': best_gene.copy(), 'best_fitness': best_fitness, 'avg_fitness': avg_fitness})
            
            if verbose:
                print('Era:             ', era)
                print('Best chromosome: ', best_gene)
                print('Best fitness:    ', best_fitness)
                print('Average fitness: ', avg_fitness)
                print('######################')

            # Select and mate
            idx = int(len(self.population)*self.retain_prob)
            parents = self.population[:idx]
            childs =  []

            while len(parents) + len(childs) < len(self.population):
                g1, g2 = random.choices(parents, k=2)
                childs.append(g1.crossover(g2))
            
            # Mutate
            for g in parents:
                g.mutate(self.mutation_prob)
            
            for g in childs:
                g.mutate(self.mutation_prob)
            
            # A new era begins
            self.population = parents + childs
            
            if target_fitness != None:
                if self.maximize and target_fitness <= best_fitness:
                    break
                elif not self.maximize and target_fitness >= best_fitness:
                    break
        return info